from datetime import datetime
from pathlib import Path

import requests_mock
from ..client import (
    from_credentials,
    from_token,
    encode_thumbnail,
    get_file_mod_time,
    get_file_size,
)

from ..model import (
    Attachment,
    Datablock,
    DataFile,
    Dataset,
    Ownable,
)

local_url = "http://localhost:3000/api/v3/"


def add_mock_requests(mock_request):
    mock_request.post(
        local_url + "Users/login",
        json={"id": "a_token"},
    )
    mock_request.post(local_url + "Samples", json={"sampleId": "dataset_id"})
    mock_request.post(local_url + "RawDatasets/replaceOrCreate", json={"pid": "42"})
    mock_request.post(
        local_url + "RawDatasets/42/origdatablocks",
        json={"response": "random"},
    )
    mock_request.post(
        local_url + "RawDatasets/42/attachments",
        json={"response": "random"},
    )


def test_scicate_ingest():
    with requests_mock.Mocker() as mock_request:
        add_mock_requests(mock_request)
        scicat = from_credentials(
            base_url=local_url,
            username="Zaphod",
            password="heartofgold",
        )
        assert (
            scicat._token == "a_token"
        ), "scicat client set the token given by the server"

        ownable = Ownable(ownerGroup="magrathea", accessGroups=["deep_though"])
        thumb_path = Path(__file__).parent / "data/SciCatLogo.png"

        time = get_file_mod_time(thumb_path)
        assert time is not None
        size = get_file_size(thumb_path)
        assert size is not None

        # RawDataset
        dataset = Dataset(
            path="/foo/bar",
            size=42,
            owner="slartibartfast",
            contactEmail="slartibartfast@magrathea.org",
            creationLocation="magrathea",
            creationTime=str(datetime.now()),
            type="raw",
            instrumentId="earth",
            proposalId="deepthought",
            dataFormat="planet",
            principalInvestigator="A. Mouse",
            sourceFolder="/foo/bar",
            scientificMetadata={"a": "field"},
            sampleId="gargleblaster",
            **ownable.dict()
        )
        dataset_id = scicat.upload_raw_dataset(dataset)

        # Datablock with DataFiles
        data_file = DataFile(path="/foo/bar", size=42)
        data_block = Datablock(
            size=42,
            version=1,
            datasetId=dataset_id,
            dataFileList=[data_file],
            **ownable.dict()
        )
        scicat.upload_datablock(data_block)

        # Attachment
        attachment = Attachment(
            datasetId=dataset_id,
            thumbnail=encode_thumbnail(thumb_path),
            caption="scattering image",
            **ownable.dict()
        )
        scicat.upload_attachment(attachment)


def test_initializers():
    with requests_mock.Mocker() as mock_request:
        add_mock_requests(mock_request)

        client = from_token(local_url, "let me in!")
        assert client._token == "let me in!"



from datetime import datetime
import enum

import base64
import hashlib
import logging
import json
from typing import List
import urllib

import requests

from pyscicat.model import Attachment, Datablock, Dataset, RawDataset, DerivedDataset

logger = logging.getLogger("splash_ingest")
can_debug = logger.isEnabledFor(logging.DEBUG)


class ScicatLoginError(Exception):
    """Represents an error encountered logging into SciCat"""

    def __init__(self, message):
        self.message = message


class ScicatCommError(Exception):
    """Represents an error encountered during communication with SciCat."""

    def __init__(self, message):
        self.message = message


class Severity(str, enum.Enum):
    warning = "warning"
    fatal = "fatal"


class ScicatClient:
    """Responsible for communicating with the Scicat Catamel server via http"""

    def __init__(
        self,
        base_url: str = None,
        token: str = False,
        username: str = None,
        password: str = None,
        timeout_seconds: int = None,
    ):
        """Initialize a new instance. This method attempts to create a tokenad_a
        from the provided username and password

        Parameters
        ----------
        base_url : str
            Base url. e.g. `http://localhost:3000/api/v3`
        username : str
            username to login with
        password : str
            password to login with
        timeout_seconds : [int], optional
            timeout in seconds to wait for http connections to return, by default None
        """
        if base_url[-1] != "/":
            base_url = base_url + "/"
        self._base_url = base_url
        self._timeout_seconds = (
            timeout_seconds  # we are hitting a transmission timeout...
        )
        self._username = username  # default username
        self._password = password  # default password
        self._token = token  # store token here
        assert self._base_url is not None, "SciCat database URL must be provided"

        logger.info(f"Starting ingestor talking to scicat at: {self._base_url}")

        if not self._token:
            assert (self._username is not None) and (
                self._password is not None
            ), "SciCat login credentials (username, password) must be provided if token is not provided"
            self._token = get_token(self._base_url, self._username, self._password)

    def _send_to_scicat(self, url, dataDict=None, cmd="post"):
        """sends a command to the SciCat API server using url and token, returns the response JSON
        Get token with the getToken method"""
        if cmd == "post":
            response = requests.post(
                url,
                params={"access_token": self._token},
                json=dataDict,
                timeout=self._timeout_seconds,
                stream=False,
                verify=True,
            )
        elif cmd == "delete":
            response = requests.delete(
                url,
                params={"access_token": self._token},
                timeout=self._timeout_seconds,
                stream=False,
            )
        elif cmd == "get":
            response = requests.get(
                url,
                params={"access_token": self._token},
                json=dataDict,
                timeout=self._timeout_seconds,
                stream=False,
            )
        elif cmd == "patch":
            response = requests.patch(
                url,
                params={"access_token": self._token},
                json=dataDict,
                timeout=self._timeout_seconds,
                stream=False,
            )
        return response

    #  Future support for samples
    # def upload_sample(self, sample):
    #     sample = {
    #         "sampleId": projected_start_doc.get('sample_id'),
    #         "owner": projected_start_doc.get('pi_name'),
    #         "description": projected_start_doc.get('sample_name'),
    #         "createdAt": datetime.isoformat(datetime.utcnow()) + "Z",
    #         "sampleCharacteristics": {},
    #         "isPublished": False,
    #         "ownerGroup": owner_group,
    #         "accessGroups": access_groups,
    #         "createdBy": self._username,
    #         "updatedBy": self._username,
    #         "updatedAt": datetime.isoformat(datetime.utcnow()) + "Z"
    #     }
    #     sample_url = f'{self._base_url}Samples'

    #     resp = self._send_to_scicat(sample_url, sample)
    #     if not resp.ok:  # can happen if sample id is a duplicate, but we can't tell that from the response
    #         err = resp.json()["error"]
    #         raise ScicatCommError(f"Error creating Sample {err}")

    def upload_dataset(self, dataset: Dataset) -> str:
        """Upload a raw or derived dataset (method is autosensing)

        Parameters
        ----------
        dataset : Dataset
            Dataset to load

        Returns
        -------
        str
            pid (or unique identifier) of the newly created dataset

        Raises
        ------
        ScicatCommError
            Raises if a non-20x message is returned
        """
        if isinstance(dataset, RawDataset):
            dataset_url = self._base_url + "RawDataSets/replaceOrCreate"
        elif isinstance(dataset, DerivedDataset):
            dataset_url = self._base_url + "DerivedDatasets/replaceOrCreate"
        else:
            logging.error(
                "Dataset type not recognized (not Derived or Raw dataset instances)"
            )
        resp = self._send_to_scicat(dataset_url, dataset.dict(exclude_none=True))
        if not resp.ok:
            err = resp.json()["error"]
            raise ScicatCommError(f"Error creating dataset {err}")
        new_pid = resp.json().get("pid")
        logger.info(f"new dataset created {new_pid}")
        return new_pid

    def upload_raw_dataset(self, dataset: Dataset) -> str:
        """Upload a raw dataset

        Parameters
        ----------
        dataset : Dataset
            Dataset to load

        Returns
        -------
        str
            pid (or unique identifier) of the newly created dataset

        Raises
        ------
        ScicatCommError
            Raises if a non-20x message is returned
        """
        raw_dataset_url = self._base_url + "RawDataSets/replaceOrCreate"
        resp = self._send_to_scicat(raw_dataset_url, dataset.dict(exclude_none=True))
        if not resp.ok:
            err = resp.json()["error"]
            raise ScicatCommError(f"Error creating raw dataset {err}")
        new_pid = resp.json().get("pid")
        logger.info(f"new dataset created {new_pid}")
        return new_pid

    def upload_derived_dataset(self, dataset: Dataset) -> str:
        """Upload a derived dataset

        Parameters
        ----------
        dataset : Dataset
            Dataset to upload

        Returns
        -------
        str
            pid (or unique identifier) of the newly created dataset

        Raises
        ------
        ScicatCommError
            Raises if a non-20x message is returned
        """
        derived_dataset_url = self._base_url + "DerivedDataSets/replaceOrCreate"
        resp = self._send_to_scicat(
            derived_dataset_url, dataset.dict(exclude_none=True)
        )
        if not resp.ok:
            err = resp.json()["error"]
            raise ScicatCommError(f"Error creating raw dataset {err}")
        new_pid = resp.json().get("pid")
        logger.info(f"new dataset created {new_pid}")
        return new_pid

    def upload_datablock(self, datablock: Datablock, datasetType: str = "RawDatasets"):
        """Upload a Datablock

        Parameters
        ----------
        datablock : Datablock
            Datablock to upload

        Raises
        ------
        ScicatCommError
            Raises if a non-20x message is returned
        """

        url = (
            self._base_url
            + f"{datasetType}/{urllib.parse.quote_plus(datablock.datasetId)}/origdatablocks"
        )
        resp = self._send_to_scicat(url, datablock.dict(exclude_none=True))
        if not resp.ok:
            err = resp.json()["error"]
            raise ScicatCommError(f"Error creating datablock. {err}")

    def upload_attachment(
        self, attachment: Attachment, datasetType: str = "RawDatasets"
    ):
        """Upload an Attachment.  Note that datasetType can be provided to determine the type of dataset
        that this attachment is attached to. This is required for creating the url that SciCat uses.

        Parameters
        ----------
        attachment : Attachment
            Attachment to upload

        datasetType : str
            Type of dataset to upload to, default is `RawDatasets`
        Raises
        ------
        ScicatCommError
            Raises if a non-20x message is returned
        """
        url = (
            self._base_url
            + f"{datasetType}/{urllib.parse.quote_plus(attachment.datasetId)}/attachments"
        )
        logging.debug(url)
        resp = requests.post(
            url,
            params={"access_token": self._token},
            timeout=self._timeout_seconds,
            stream=False,
            json=attachment.dict(exclude_none=True),
            verify=True,
        )
        if not resp.ok:
            err = resp.json()["error"]
            raise ScicatCommError(f"Error  uploading thumbnail. {err}")

    def get_datasets_full_query(self, skip=0, limit=25, query_fields=None):
        """Gets datasets using the fullQuery mechanism of SciCat. This is
        appropriate for cases where might want paging and cases where you want to perform
        a text search on the Datasets collection. The full features of fullQuery search
        are beyond this document.

        There is no known mechanism to query for fields that are missing or contain a
        a null value.

        To query based on the full text search, send `{"text": "<text to query"}` as query field

        Parameters
        ----------
        skip : int
            number of items to skip

        limit : int
            number of items to return

        query_fields : dict
            dictionary of terms to send to the query (must be json serializable)

        """
        if not query_fields:
            query_fields = {}
        query_fields = json.dumps(query_fields)
        query = f'fields={query_fields}&limits={{"skip":{skip},"limit":{limit},"order":"creationTime:desc"}}'

        url = f"{self._base_url}/Datasets/fullquery?{query}"
        response = self._send_to_scicat(url, cmd="get")
        if not response.ok:
            err = response.json()["error"]
            logger.error(f'{err["name"]}, {err["statusCode"]}: {err["message"]}')
            return None
        return response.json()

    def get_datasets(self, filter_fields=None) -> List[Dataset]:
        """Gets datasets using the simple fiter mechanism. This
        is appropriate when you do not require paging or text search, but
        want to be able to limit results based on items in the Dataset object.

        For example, a search for Datasets of a given proposalId would have
        ```python
        filterField = {"proposalId": "1234"}
        ```
        A search for Datasets  with no proposalId would be:
        ```python
        filterField = {"proposalId": ""}
        ```

        Parameters
        ----------
        filter_fields : dict
            Dictionary of filtering fields. Must be json serializable.
        """
        if not filter_fields:
            filter_fields = {}

        filter_fields = json.dumps(filter_fields)
        url = f'{self._base_url}/Datasets/?filter={{"where":{filter_fields}}}'
        response = self._send_to_scicat(url, cmd="get")
        if not response.ok:
            err = response.json()["error"]
            logger.error(f'{err["name"]}, {err["statusCode"]}: {err["message"]}')
            return None
        return response.json()


def get_file_size(pathobj):
    filesize = pathobj.lstat().st_size
    return filesize


def get_checksum(pathobj):
    with open(pathobj) as file_to_check:
        # pipe contents of the file through
        return hashlib.md5(file_to_check.read()).hexdigest()


def encode_thumbnail(filename, imType="jpg"):
    logging.info(f"Creating thumbnail for dataset: {filename}")
    header = "data:image/{imType};base64,".format(imType=imType)
    with open(filename, "rb") as f:
        data = f.read()
    dataBytes = base64.b64encode(data)
    dataStr = dataBytes.decode("UTF-8")
    return header + dataStr


def get_file_mod_time(pathobj):
    # may only work on WindowsPath objects...
    # timestamp = pathobj.lstat().st_mtime
    return str(datetime.fromtimestamp(pathobj.lstat().st_mtime))


def from_token(base_url: str, token: str):
    return ScicatClient(base_url, token)


def from_credentials(base_url: str, username: str, password: str):
    token = get_token(base_url, username, password)
    return from_token(base_url, token)


def get_token(base_url, username, password):
    """logs in using the provided username / password combination
    and receives token for further communication use"""
    logger.info(f" Getting new token for user {username}")
    if base_url[-1] != "/":
        base_url = base_url + "/"
    response = requests.post(
        base_url + "Users/login",
        json={"username": username, "password": password},
        stream=False,
        verify=True,
    )
    if not response.ok:
        logger.error(f" ** Error received: {response}")
        err = response.json()["error"]
        logger.error(f' {err["name"]}, {err["statusCode"]}: {err["message"]}')
        raise ScicatLoginError(response.content)

    data = response.json()
    # print("Response:", data)
    token = data["id"]  # not sure if semantically correct
    logger.info(f" token: {token}")
    return token



# coding: utf-8
# h5tools: convenience functions for dealing with hdf5 and nexus files
# author: Sofya Laskina, Brian R. Pauw, I. Bressler
# date: 2022.01.10

import h5py
import logging
import numpy as np
from pathlib import Path


def h5Get(filename, h5path: str = None, default="none", leaveAsArray=False):
    """
    Gets a single value or attribute from an HDF5 file, with added error checking and default handling.
    h5path is the string representation of the location in the hdf5 file, e.g. '/sasentry1/sasdata1/I'.
    If you want to extract an attribute, you can use the '@' symbol to split path and attribute name,
    e.g. '/sasentry1/sasdata1@timestamp' gets the timestamp attribute in sasdata1.
    """
    assert h5path is not None, "h5path must be specified"
    assert isinstance(h5path, str), "h5path must be a string"
    filename = Path(filename)
    assert filename.exists(), f"input filename {filename.as_posix()} cannot be found"
    attrKey = None
    if "@" in h5path:
        h5path, attrKey = h5path.split("@")

    with h5py.File(filename, "r") as h5f:
        try:
            if attrKey is None:
                val = h5f.get(h5path)[()]
            else:
                val = h5f.get(h5path).attrs[attrKey]

            val = h5py_casting(val)  # sofya added this line
            # logging.info('type val {} at key {}: {}'.format(val, h5path, type(val)))

        except TypeError:
            if attrKey is None:
                logging.warning(
                    f"cannot get value from file {filename.as_posix()} path {h5path}, setting to default"
                )
            else:
                logging.warning(
                    f"""cannot get value from file {filename.as_posix()} path {h5path},
                    attribute {attrKey} setting to default: {default}"""
                )

            val = default
    return val


def h5GetDict(filename, keyPaths: dict):
    """
    creates a dictionary with results extracted from an HDF5 file
    dictionary should have form:
    {h5path: default}
    """
    resultDict = {}
    for h5path, default in keyPaths.items():
        resultDict[h5path] = h5Get(filename, h5path, default=default)
    return resultDict


def h5py_casting(val, leaveAsArray=False):
    if isinstance(val, np.ndarray) and (not leaveAsArray):
        if val.size == 1:
            val = np.array([val.squeeze()])[0]
        else:
            if np.isnan(val).sum() + np.isinf(val).sum() == np.prod(
                [i for i in val.shape]
            ):
                # print('all elements are either nan or inf')
                val = "-"
            elif np.isnan(val.mean()) or np.isinf(val.mean()):
                # print('nan pixel at index', np.argwhere(np.isnan(val)))
                # print('inf pixel at index', np.argwhere(np.isinf(val)))
                val = np.mean(np.ma.masked_invalid(val))
            else:
                val = val.mean()
    """if isinstance( val, np.ndarray) and leaveAsArray: seems to take a lot of time
        val = val.tolist()
        return json.dumps(val, separators=(',', ':'), sort_keys=True, indent=4)"""
    if isinstance(val, float) and (np.isnan(val) or np.isinf(val)):
        val = "-"
    if isinstance(val, np.bytes_) or isinstance(val, bytes):
        val = val.decode("UTF-8")
    if isinstance(val, np.generic):
        val = val.item()
    if isinstance(val, str):
        if val[:2] == "b'":
            val = val[2:-1]
    return val



# coding: utf-8
# sciMeta: extract all metadata from a NeXus file and put it in a tree for upload as scientific metadata to SciCat
# author: Sofya Laskina, Brian R. Pauw
# date: 2022.01.10

# from unittest import skip # not sure where this import comes from
import h5py

# flake8: noqa: F401
import hdf5plugin  # ESRF's library that extends the read functionality of HDF5 files
from pyscicat.hdf5.h5tools import h5py_casting
import logging
from pathlib import Path
from collections import abc


def update_deep(dictionary: dict, path_update: dict) -> dict:
    """
    Update the main metadata dictionary with the new dictionary.
    """
    k = list(path_update.keys())[0]
    v = list(path_update.values())[0]
    if k not in dictionary.keys():
        dictionary[k] = v
    else:
        key_next = list(path_update[k].keys())[0]
        if key_next in dictionary[k].keys():
            dictionary[k] = update_deep(dictionary.get(k, {}), v)
        else:
            dictionary[k].update(v)
    return dictionary


def build_dictionary(levels, update_data: dict) -> dict:
    """
    Creates a json-like level based dictionary for the whole path starting from /entry1 or whatever the
    first child of the root in the datatree is.
    """
    for level in levels[::-1]:
        update_data = dict({level: update_data})
    return update_data


def unwind(
    h5f,
    parent_path,
    metadata,
    default="none",
    leaveAsArray=False,
    skipKeyList: list = [],
) -> dict:
    """
    Current_level is the operating level, that is one level higher that the collected data.
    """
    if isinstance(h5f.get(parent_path), abc.Mapping):
        new_keys = sorted(h5f.get(parent_path).keys())
        # remove unwanted items before we even get started on it
        keyList = [newKey for newKey in new_keys if newKey not in skipKeyList]

        for nk in keyList:
            unwind(h5f, "/".join([parent_path, nk]), metadata, skipKeyList=skipKeyList)
    else:
        try:
            val = h5f.get(parent_path)[()]
            val = h5py_casting(val, leaveAsArray)
        except (OSError, TypeError):
            logging.warning(
                f"file has no value at path {parent_path}, setting to default: {default}"
            )
            val = default

        attributes = {"value": val}
        try:
            attributes_add = h5f.get(parent_path).attrs
            a_key = attributes_add.keys()
            a_value = []
            for v in attributes_add.values():
                v = h5py_casting(v, leaveAsArray)
                a_value.append(v)
            attributes.update(dict(zip(a_key, a_value)))
        except (KeyError, AttributeError) as e:
            logging.warning(e)

        levels = parent_path.split("/")[1:]
        if list(attributes.keys()) == ["value"]:  # no attributes here
            nested_dict = val
        else:
            nested_dict = attributes.copy()
        if val != "":
            update_dict = build_dictionary(levels, nested_dict)
            metadata = update_deep(metadata, update_dict)


def scientific_metadata(
    filename, excludeRootEntry: bool = True, skipKeyList: list = []
) -> dict:
    """
    Goals:
    --
    Opens any HDF5 or nexus file and unwinds the structure to add up all the metadata and respective attributes.
    This adds the paths and structure as required for SciCat's "scientific metadata" upload, including units.

    Usage:
    --
    branches and keys to omit can be listed using the argument "skipKeyList". Example:
    scientificMetadata=scientific_metadata(Path('./my_file.h5'), skipKeyList=['sasdata1'])
    If the root branch is singular, it can be omitted from the output dictionary by setting
    excludeRootEntry to True

    """
    # ensure the filename argument is of class Path
    filename = Path(filename)
    assert (
        filename.exists()
    ), f"Input filename {filename.as_posix()} does not seem to exist."
    with h5py.File(filename, "r") as h5f:
        # let's see if we can do this simpler
        metadata = dict()  # .fromkeys(prior_keys)
        unwind(h5f, "/", metadata, skipKeyList=skipKeyList)

    # first metadata entry is empty, so enter one level deeper.
    if len(metadata.keys()) == 1 and list(metadata.keys())[0] == "":
        metadata = metadata[list(metadata.keys())[0]]

    if excludeRootEntry and (len(metadata.keys()) > 1):
        logging.warning(
            """root entry cannot be excluded when there are more than one in the HDF5 tree.
            excludeRootEntry flag will be ignored."""
        )
    if excludeRootEntry and (len(metadata.keys()) == 1):
        metadata = metadata[list(metadata.keys())[0]]

    return metadata



from pathlib import Path
from setuptools import setup, find_packages
import sys
import versioneer

min_version = (3, 7)
if sys.version_info < min_version:
    error = """
pyscicat does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(
        *(sys.version_info[:2] + min_version)
    )
    sys.exit(error)

here = Path(__file__).absolute()

with open(here.with_name("README.md"), encoding="utf-8") as readme_file:
    readme = readme_file.read()


def read_requirements_from_here(here: Path, filename: str = None) -> list:
    assert filename is not None, "filename as string must be provided"
    assert here.with_name(
        filename
    ).exists(), f"requirements filename {filename.as_posix()} does not exist"
    with open(here.with_name(filename)) as requirements_file:
        # Parse requirements.txt, ignoring any commented-out lines.
        requirements = [
            line
            for line in requirements_file.read().splitlines()
            if not line.startswith("#")
        ]
    return requirements


extras_require = {}
extras_require["base"] = read_requirements_from_here(here, "requirements.txt")
extras_require["h5tools"] = read_requirements_from_here(here, "requirements-hdf5.txt")

setup(
    name="pyscicat",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Code for communicating to a SciCat backend server python",
    long_description=readme,
    author="Dylan McReynolds",
    author_email="dmcreynolds@lbl.gov",
    url="https://github.com/scicatproject/pyscicat",
    python_requires=">={}".format(".".join(str(n) for n in min_version)),
    packages=find_packages(exclude=["docs", "tests"]),
    include_package_data=True,
    extras_require=extras_require,
    install_requires=extras_require["base"],
    license="BSD (3-clause)",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
)



import enum

# from re import L
from typing import List, Dict, Optional

from pydantic import BaseModel


class DatasetType(str, enum.Enum):
    """type of Dataset"""

    raw = "raw"
    derived = "derived"


class Ownable(BaseModel):
    """Many objects in SciCat are ownable"""

    ownerGroup: str
    accessGroups: List[str]


class MongoQueryable(BaseModel):
    """Many objects in SciCat are mongo queryable"""

    createdBy: Optional[str]
    updatedBy: Optional[str]
    updatedAt: Optional[str]
    createdAt: Optional[str]


class User(BaseModel):
    """Base user."""

    # TODO: find out which of these are not optional and update
    realm: str
    username: str
    email: str
    emailVerified: bool = False
    id: str


class Proposal(Ownable, MongoQueryable):
    """
    Defines the purpose of an experiment and links an experiment to principal investigator and main proposer
    """

    # TODO: find out which of these are not optional and update
    proposalId: Optional[str]
    pi_email: Optional[str]
    pi_firstname: Optional[str]
    pi_lastname: Optional[str]
    email: Optional[str]
    firstname: Optional[str]
    lastname: Optional[str]
    title: Optional[str]
    abstract: Optional[str]
    startTime: Optional[str]
    endTime: Optional[str]
    MeasurementPeriodList: Optional[
        List[dict]
    ]  # may need updating with the measurement period model


class Sample(Ownable, MongoQueryable):
    """
    Models describing the characteristics of the samples to be investigated.
    Raw datasets should be linked to such sample definitions.
    """

    # TODO: find out which of these are not optional and update
    sampleId: Optional[str]
    owner: Optional[str]
    description: Optional[str]
    sampleCharacteristics: Optional[dict]
    isPublished: bool = False


class Job(MongoQueryable):
    """
    This collection keeps information about jobs to be excuted in external systems.
    In particular it keeps information about the jobs submitted for archiving or
    retrieving datasets stored inside an archive system. It can also be used to keep
    track of analysis jobs e.g. for automated analysis workflows
    """

    id: Optional[str]
    emailJobInitiator: str
    type: str
    creationTime: Optional[str]  # not sure yet which ones are optional or not.
    executionTime: Optional[str]
    jobParams: Optional[dict]
    jobStatusMessage: Optional[str]
    datasetList: Optional[dict]  # documentation says dict, but should maybe be list?
    jobResultObject: Optional[dict]  # ibid.


class Instrument(MongoQueryable):
    """
    Instrument class, most of this is flexibly definable in customMetadata
    """

    pid: Optional[str]
    name: str
    customMetadata: Optional[dict]


class Dataset(Ownable, MongoQueryable):
    """
    A dataset in SciCat, base class for derived and raw datasets
    """

    pid: Optional[str]
    classification: Optional[str]
    contactEmail: str
    creationTime: str  # datetime
    datasetName: Optional[str]
    description: Optional[str]
    history: Optional[List[dict]]  # list of foreigh key ids to the Messages table
    instrumentId: Optional[str]
    isPublished: Optional[bool] = False
    keywords: Optional[List[str]]
    license: Optional[str]
    numberOfFiles: Optional[int]
    numberOfFilesArchived: Optional[int]
    orcidOfOwner: Optional[str]
    packedSize: Optional[int]
    owner: str
    ownerEmail: Optional[str]
    sharedWith: Optional[List[str]]
    size: Optional[int]
    sourceFolder: str
    sourceFolderHost: Optional[str]
    techniques: Optional[List[dict]]  # with {'pid':pid, 'name': name} as entries
    type: DatasetType
    validationStatus: Optional[str]
    version: Optional[str]


class RawDataset(Dataset):
    """
    Raw datasets from which derived datasets are... derived.
    """

    principalInvestigator: Optional[str]
    creationLocation: Optional[str]
    dataFormat: str
    type: DatasetType = "raw"
    createdAt: Optional[str]  # datetime
    updatedAt: Optional[str]  # datetime
    dataFormat: Optional[str]
    endTime: Optional[str]  # datetime
    sampleId: Optional[str]
    proposalId: Optional[str]
    scientificMetadata: Optional[Dict]


class DerivedDataset(Dataset):
    """
    Derived datasets which have been generated based on one or more raw datasets
    """

    investigator: Optional[str]
    inputDatasets: List[str]
    usedSoftware: List[str]  # not optional!
    jobParameters: Optional[dict]
    jobLogData: Optional[str]
    scientificMetadata: Optional[Dict]


class DataFile(MongoQueryable):
    """
    A reference to a file in SciCat. Path is relative
    to the Dataset's sourceFolder parameter

    """

    path: str
    size: int
    time: Optional[str]
    uid: Optional[str] = None
    gid: Optional[str] = None
    perm: Optional[str] = None


class Datablock(Ownable):
    """
    A Datablock maps between a Dataset and contains DataFiles
    """

    id: Optional[str]
    # archiveId: str = None  listed in catamel model, but comes back invalid?
    size: int
    packedSize: Optional[int]
    chkAlg: Optional[int]
    version: str = None
    dataFileList: List[DataFile]
    datasetId: str


class Attachment(Ownable):
    """
    Attachments can be any base64 encoded string...thumbnails are attachments
    """

    id: Optional[str]
    thumbnail: str
    caption: Optional[str]
    datasetId: str
